package org.openjdk.gcbench.fragger;

import org.openjdk.gcbench.util.AllocProfileSupport;
import org.openjdk.gcbench.util.ratelimit.MultiTokenBucket;
import org.openjdk.jmh.annotations.*;
import org.openjdk.jol.info.GraphLayout;

import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.TimeUnit;

@Warmup(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Measurement(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Fork(value = 1)
@BenchmarkMode(Mode.Throughput)
@OutputTimeUnit(TimeUnit.SECONDS)
@Threads(Threads.MAX)
@State(Scope.Benchmark)
public class ArrayFragger {

    @Param("16")
    int payloadSize;

    @Param({"1000"})
    long ldsMB;

    @Param({"10"})
    int rate;

    Object[] objects;

    int count;

    MultiTokenBucket bucket;

    Object[] o0;
    Object[] o16;

    private int getSizePerCount() {
        if (AllocProfileSupport.isAvailable()) {
            long a1 = AllocProfileSupport.getAllocatedBytes();
            o0 = new Object[0];
            long a2 = AllocProfileSupport.getAllocatedBytes();
            o16 = new Object[16];
            for (int c = 0; c < 16; c++) {
                o16[c] = new byte[payloadSize];
            }
            long a3 = AllocProfileSupport.getAllocatedBytes();

            long sizeO0 = a2 - a1;
            long sizeO16 = a3 - a2;

            return (int) ((sizeO16 - sizeO0) / 16);
        } else {
            // best guess
            return 4 + // array element
                    16 + payloadSize; // object
        }
    }

    @Setup
    public void setup() {
        bucket = new MultiTokenBucket(rate);
        count = (int) Math.max(1, ldsMB * 1024 * 1024 / getSizePerCount());
        objects = new Object[count];
        for (int c = 0; c < count; c++) {
            objects[c] = new byte[payloadSize];
        }
    }

    @Benchmark
    public void test() {
        bucket.limit();
        objects[ThreadLocalRandom.current().nextInt(count)] = new byte[payloadSize];
    }

}
